/*
**                       <|A| |S| |S| |I| |T| |C| |H|>
**
** Purpose:       AssItch analyses Packet Filter Rules for incoming and
**                outgoing TCP Services
**
** Version:       2.6
**
** Last Update:   2000/08/24
**
** Author:        Thomas Biege (thomas@suse.de)
**
** Tested on:     Linux 2.0.3x (You have to use the netinet/tcp.h file of 2.2.x)
**                Linux 2.2.13 (SuSE Linux 6.3)
**
** The GPL applies to this source code.
**
*/

#include <string.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <unistd.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/ip_icmp.h>
#include <netinet/tcp.h>

#include "assitch.h"
#include "assitch-defs.h"
#include "error.h"
#include "rtt.h"
#include "dns_lib.h"

extern int errno;
FILE *output;

  
/*
**               M A I N
*/
int main(int argc, char **argv)
{
  u_long          saddr, daddr;
  u_short         priv_port = 0, npriv_port = 0, scan;
  u_short         *scnprts = NULL;
  int             on = 1;
  u_int           port, opt;
  enum scans      scanmethod;
	char 						outputfile[NAME_MAX+1];


  prog = argv[0];
	output = stderr;

  opterr = 0;
  while((opt = getopt(argc, argv, "o:p:u:hf")) != EOF)
    switch(opt)
    {
			case 'o':	if(optarg == NULL || optarg[0] == '-')
									USAGE(prog)
								strncpy(outputfile, optarg, sizeof(outputfile)-1)[sizeof(outputfile)-1] = '\0';
								if((output = fopen(outputfile, "w")) == NULL)
									err_mesg(FATAL_SYS, "ERROR: fopen()\nSYSERR");
								break;
      case 'p': priv_port = (u_short) atoi(optarg);
                break;
      case 'u': npriv_port = (u_short) atoi(optarg);
                break;
      case 'h': htmlflag = 1;
                break;
      case 'f': finflag = 1;
                break;
      default:  USAGE(prog)
    }

  argc -= optind;
  argv += optind;

  if(argc != 3)
    USAGE(prog)
  
  /* Set and check Parameters from Commandline */
  saddr   = nameResolve(argv[0]);
  daddr   = nameResolve(argv[1]);

  if(!isdigit(argv[2][0]))
    USAGE(prog)

  scnprts = getpts(argv[2]);
  
  if(priv_port)
    ports[0] = priv_port;
  if(npriv_port)
    ports[1] = npriv_port;

  if(set_signal(SIGALRM, rtt_alarm) < 0)
    err_mesg(FATAL_SYS, "ERROR: set_siganl()\nSYSERR");

  /*
  ** Open Sockets
  */
  /* raw socket for sending */
  if( (send_sock = socket(AF_INET, SOCK_RAW, IPPROTO_RAW)) < 0 )
    err_mesg(FATAL_SYS, "Can't create raw IP send-socket.\nSYSERR");
  if( setsockopt(send_sock, IPPROTO_IP, IP_HDRINCL, (char *) &on, sizeof(on)) < 0 )
    err_mesg(FATAL_SYS, "Can't set socket option IP_HDRINCL.\nSYSERR");
            
  /* raw tcp socket for receiving */
  if( (recv_sock = socket(AF_INET, SOCK_RAW, IPPROTO_TCP)) < 0 )
    err_mesg(FATAL_SYS, "Can't create raw TCP recv-socket.\nSYSERR");

  /* icmp socket to detect icmp unreachable messages */
  if( (icmp_sock = socket(AF_INET, SOCK_RAW, IPPROTO_ICMP)) < 0 )
    err_mesg(FATAL_SYS, "Can't create ICMP socket\nSYSERR");
                                        
  
  /*
  ** Set up Pseudo Header
  */
  p_hdr           = (struct pseudohdr *) ChkSumBuf;
  p_hdr->saddr    = saddr;
  p_hdr->daddr    = daddr;
  p_hdr->pad      = 0;
  p_hdr->proto    = IPPROTO_TCP;
  p_hdr->tcplen   = htons(TCPHDR);
                  

  if(htmlflag)
  {
    HTMLHDR(hostLookup(daddr))
    TABLHDR
  }
  else
  {
    HDR(hostLookup(daddr))
  }
                         
  /*
  ** Let's start Scanning
  */
  for(scan = 0; scnprts[scan] != 0; scan++)
  {
    for(port = 0; port < MAXPORTS; port++)
    {
      for(scanmethod = finflag ? finscan : synscan; scanmethod != stopscan; scanmethod--)
        if(chi(saddr, daddr, htons(ports[port]), htons(scnprts[scan]), scanmethod) < 0)
          err_mesg(WARN, "ERROR: chi()\n");
            
      if(htmlflag)
        TABLSPACE
      else
        fprintf(output, "|\n");
    }
    if(htmlflag)
      TABLSPACE
    else
      fprintf(output, "|-------------------------------------------------------------------------------\n");
  }
    
  if(htmlflag)
    TABLCLOSE
  else
    fprintf(output, "+-------------------------------------------------------------------------------\n");
        
	if(output != stderr)
		fclose(output);
  exit(0);
}

/*
** chi()
*/
int chi(u_long saddr, u_long daddr, u_short defport, u_short port, enum scans mode)
{
  char                sendpack[TCPIPHDR] = { 0 };
  char                recvpack[MAXPACK]  = { 0 };
  struct iphdr        *ip;
  struct tcphdr       *tcp;
  int                 maxfd;
  int                 nread;
  fd_set              readset;
  sigset_t            newsigset, oldsigset;


  if(mode != ackscan && mode != synscan && mode != finscan)
    err_mesg(FATAL, "chi(): mode sux\n");


  if( rttfirst )
  {
    rtt_init(&rttinfo);   /* init. first time we are called */
    rttfirst = 0;
  }

  sigemptyset(&newsigset);
  sigaddset(&newsigset, SIGALRM);

  
  /*
  ** Set up Socket Info
  */
  sockinfo.sin_family       = AF_INET;
  sockinfo.sin_port         = (mode == ackscan) ? defport : port;
  sockinfo.sin_addr.s_addr  = daddr;
  sockinfolen = sizeof(sockinfo);
        
  /*
  ** Set up IP Header
  */
  ip = (struct iphdr *) sendpack;
  ip->saddr     = saddr;
  ip->daddr     = daddr;
  ip->version   = 4;
  ip->ihl       = 5;
  ip->ttl       = 0xFF;
  ip->protocol  = IPPROTO_TCP;
  ip->tot_len   = htons(IPHDR);
  
  /*
  ** Set up TCP Header
  */
  tcp = (struct tcphdr *) (sendpack + IPHDR);
  if(mode == ackscan)
  {
    tcp->source   = port;
    tcp->dest     = defport;
    tcp->ack      = 1;
    tcp->ack_seq  = htonl((u_long) 31337);  /* Take care! ;) */
  }
  else if(mode == synscan)
  {
    tcp->source = defport;
    tcp->dest   = port;
    tcp->seq    = 1;
  }
  else
  {
    tcp->source = defport;
    tcp->dest   = port;
    tcp->fin    = 1;
  }
  tcp->seq  = htonl((u_long) rand());   /* Who cares? */
  tcp->doff = 5;
              
  /*
  ** Calculate the Checksum
  */
  bcopy(tcp, ChkSumBuf+PHDR, TCPHDR);
  tcp->check   = in_chksum((u_short *) ChkSumBuf, (int) PTCPHDR);


  rtt_newpack(&rttinfo);  /* init. for new packet */

rexmit:
  if(sendto(send_sock, sendpack, TCPIPHDR, 0, (struct sockaddr *) &sockinfo, sockinfolen) < 0)
    err_mesg(FATAL_SYS, "ERROR: sendto()\nSYSERR");
    
  delay(DELAY);

  /*
  ** Set RTT and select() Stuff.
  */
  errno = rtt_to = 0;             /* for signal handler */
  alarm( rtt_start(&rttinfo) );   /* calc. timeout value & start timer */
  
  FD_ZERO(&readset);
  maxfd = (icmp_sock > recv_sock) ? (icmp_sock + 1) : (recv_sock + 1);

  while(1)
  {
    FD_SET(recv_sock, &readset);
    FD_SET(icmp_sock, &readset);
        
    if(select(maxfd, &readset, NULL, NULL, NULL) < 0 )
    {
      if(rtt_to)
      {
        /*
        ** The select() timed out.
        ** If we have tried enough, then let's quit.
        */
        if( rtt_timeout(&rttinfo) < 0 )
        {
          rttfirst = 1;
          if(htmlflag)
            HTMLDENY
          else
            DENY
                      
          return(1);
        }

        /*
        ** We have to send the packet again.
        */
#ifdef VERBOSE
        err_mesg(WARN, "Retransmission!\n");
#endif
        goto rexmit;
      }
      
#ifdef VERBOSE
        err_mesg(WARN_SYS, "ERROR: select()\nSYSERR");
#endif
      continue;
    }

    if(sigprocmask(SIG_BLOCK, &newsigset, &oldsigset) < 0)
      err_mesg(FATAL_SYS, "ERROR: sigprocmask(BLOCK)\nSYSERR");
         
    /*
    ** Either we cought a TCP packet or a ICMP packet
    */
    
    if(FD_ISSET(recv_sock, &readset))
    {
      /*
      ** We cought a TCP pack, so let's read it and
      ** pass it to check_tcp()
      */

      if((nread = read(recv_sock, recvpack, MAXPACK)) < 0)
        err_mesg(FATAL_SYS, "ERROR: read()\nSYSERR");
        
      if(check_tcp(sendpack, recvpack, nread, mode) > 0)
      {
        alarm(0);           /* stop signal timer */
        rtt_stop(&rttinfo); /* stop RTT timer, calc & store new values */
        break;
      }
#ifdef VERBOSE
      else
        err_mesg(WARN, "ERROR: check_tcp()\n");
#endif
        
    }
    
    if(FD_ISSET(icmp_sock, &readset))
    {
      /*
      ** We cought a ICMP pack, so let's read + check it and
      ** pass it to check_icmp()
      */
      if((nread = read(icmp_sock, recvpack, MAXPACK)) < 0)
        err_mesg(FATAL_SYS, "ERROR: read()\nSYSERR");
      
      if(check_icmp(sendpack, recvpack, nread, mode) > 0)
      {
        alarm(0);           /* stop signal timer */
        rtt_stop(&rttinfo); /* stop RTT timer, calc & store new values */
        break;
      }
#ifdef VERBOSE
      else
        err_mesg(WARN, "ERROR: check_icmp()\n");
#endif
    }

    memset(recvpack, 0, MAXPACK);
  
    if(sigprocmask(SIG_SETMASK, &oldsigset, NULL) < 0)
      err_mesg(FATAL_SYS, "ERROR: sigprocmask(SETMASK)\nSYSERR");

  }   /* while(1) */

  if(sigprocmask(SIG_SETMASK, &oldsigset, NULL) < 0)
    err_mesg(FATAL_SYS, "ERROR: sigprocmask(SETMASK)\nSYSERR");

  return(0);
}


/*
** check_tcp() checks TCP packets
** return values: -1  : error
**                 1  : everything is in order
**                 0  : something is wrong
*/
int check_tcp(char *sendpack, char *recvpack, int nread, u_int mode)
{
  struct iphdr  *ip_recv;
  struct iphdr  *ip_send;
  struct tcphdr *tcp_recv;
  struct tcphdr *tcp_send;
  u_int         iphdrlen;
  u_int         syn_flag = 0, rst_flag = 0, fin_rst_flag = 0;

  
  ip_recv = (struct iphdr *) recvpack;

  iphdrlen = ip_recv->ihl << 2;
  if(nread < (iphdrlen + TCPHDR))
  {
#ifdef VERBOSE
    err_mesg(WARN, "TCP packet is too short!\n");
    fflush(stdout);
#endif
    return(0);  /* packet too short */
  }
  
  tcp_recv = (struct tcphdr *) (recvpack + iphdrlen);
  ip_send = (struct iphdr *) sendpack;
  tcp_send = (struct tcphdr *) (sendpack + IPHDR);
  
    
  if( ip_recv->saddr     != ip_send->daddr     ||
      tcp_recv->source != tcp_send->dest ||
      tcp_recv->dest != tcp_send->source
     )
  {
#ifdef VERBOSE
    err_mesg(WARN, "TCP packet doesn't belong to us!\n");
    fflush(stdout);
#endif
    return(0);
  }
                              
  
  switch(mode)
  {
    case ackscan: if( (tcp_recv->rst) && (tcp_recv->seq == tcp_send->ack_seq) )
                    rst_flag = 1;
                  break;
    case synscan: if(tcp_recv->rst)
                    rst_flag = 1;
                  else if(tcp_recv->syn && tcp_recv->ack)
                    syn_flag = 1;
                  else
                    return(-1); /* change defport! */
                  break;
    case finscan: if(tcp_recv->rst)
                    fin_rst_flag = 1;
                  break;
    default:      return(-1);
                  /* change defport!... hey, or we may hit the right seq# ;) */
  }
  
  if(htmlflag)
    HTMLRESULT
  else
    RESULT

   
  /*
  ** We don't have to send a RST packet, because the TCP stack will
  ** do it for us.
  */

  /*
  ** We should send an RST packet if we received an SYN|ACK packet
  ** to avoid a DoS attack.
  if(syn_flag)
  {
    tcp_send->seq     = tcp_recv->ack_seq;
    tcp_send->ack_seq = htonl( (u_long) (ntohl(tcp_recv->seq) + 1) );
    tcp_send->rst     = 1;
    
    bcopy(tcp_send, ChkSumBuf+PHDR, TCPHDR);
    tcp_send->check   = in_chksum((u_short *) ChkSumBuf, (int) PTCPHDR);
    
    if(sendto(send_sock, sendpack, TCPIPHDR, 0, (struct sockaddr *) &sockinfo, sockinfolen) < 0)
      err_mesg(FATAL_SYS, "ERROR: sendto()\nSYSERR");
  }
  */
  
  return(1);
}

/*
** check_icmp() checks ICMP packets
** return values: -1  : error
**                 1  : everything is in order
**                 0  : something is wrong
*/
int check_icmp(char *sendpack, char *recvpack, int nread, u_int mode)
{
  struct iphdr    *ip_send;
  struct iphdr    *ip_recv;
  struct iphdr    *my_ip;
  struct icmphdr  *icmp;
  struct tcphdr   *tcp_send;
  struct tcphdr   *my_tcp;
  u_int           iphdrlen;
  u_int           port_ur = 0, prohib = 0;
  

  ip_send = (struct iphdr *) sendpack;  
  ip_recv = (struct iphdr *) recvpack;
  
  iphdrlen = ip_recv->ihl << 2;
  if(nread < (iphdrlen + ICMPHDR))
  {
#ifdef VERBOSE
    err_mesg(WARN, "ICMP packet is too short!\n");
    fflush(stdout);
#endif
    return(0);  /* packet too short */
  }
  
  icmp = (struct icmphdr *) (recvpack + iphdrlen);
  my_ip = (struct iphdr *) (recvpack + iphdrlen + ICMPHDR);
  my_tcp = (struct tcphdr *) (recvpack + iphdrlen + ICMPHDR + IPHDR);
  tcp_send = (struct tcphdr *) (sendpack + IPHDR);

  if( my_ip->daddr     != ip_send->daddr      ||
      my_tcp->source != tcp_send->source  ||
      my_tcp->dest != tcp_send->dest
    )
  {
#ifdef VERBOSE
    err_mesg(WARN, "ICMP packet doesn't belong to us!\n");
    fflush(stdout);
#endif
    return(0);
  }
            
    
  if(icmp->type == ICMP_DEST_UNREACH)
  {
    if(icmp->code == ICMP_PORT_UNREACH)
      port_ur = 1;
    else if(icmp->code == ICMP_PROHIBITED)
      prohib = 1;
    else
      return(1);
  }
  else
  {
#ifdef VERBOSE
    err_mesg(WARN, "Not a ICMP unreachable message!\n");
#endif
    return(0);
  }
  
      
  if(htmlflag)
    HTMLICMP
  else
    PRTICMP
    
  return(1);
}


u_short in_chksum(u_short *ptr, int nbytes)
{
  register long           sum;            /* assumes long == 32 bits */
  u_short                 oddbyte;
  register u_short        answer;         /* assumes u_short == 16 bits */
  
  /*
  * Our algorithm is simple, using a 32-bit accumulator (sum),
  * we add sequential 16-bit words to it, and at the end, fold back
  * all the carry bits from the top 16 bits into the lower 16 bits.
  */

  sum = 0;
  while (nbytes > 1)
  {
    sum += *ptr++;
    nbytes -= 2;
  }

  /* mop up an odd byte, if necessary */
  if (nbytes == 1)
  {
    oddbyte = 0;            /* make sure top half is zero */
    *((unsigned char *) &oddbyte) = *(unsigned char *) ptr;   /* one byte only */
    sum += oddbyte;
  }

  /*
   * Add back carry outs from top 16 bits to low 16 bits.
   */

  sum  = (sum >> 16) + (sum & 0xffff);    /* add high-16 to low-16 */
  sum += (sum >> 16);                     /* add carry */
  answer = ~sum;          /* ones-complement, then truncate to 16 bits */

  return((u_short) answer);
}

u_short *getpts(char *origexpr)
{
  int exlen = strlen(origexpr);
  char *p,*q;
  unsigned short *tmp, *ports;
  int i=0, j=0,start,end;
  char *expr = strdup(origexpr);
  char *mem = expr;

  ports = (u_short *) malloc(65536 * sizeof(short));

  for(;j < exlen; j++) 
    if (expr[j] != ' ') expr[i++] = expr[j]; 

  expr[i] = '\0';
  exlen = i;
  i=0;

  while((p = strchr(expr,',')))
  {
    *p = '\0';
    if (*expr == '-')
    {
      start = 1;
      end = atoi(expr+ 1);
    }
    else
    {
      start = end = atoi(expr);
      if ((q = strchr(expr,'-')) && *(q+1) )
	end = atoi(q + 1);
      else if (q && !*(q+1))
	end = 65535;
    }
  
    if (start < 1 || start > end)
      err_mesg(FATAL, "Your port specifications are illegal!");

    for(j=start; j <= end; j++) 
      ports[i++] = j;
    expr = p + 1;
  }

  if (*expr == '-')
  {
    start = 1;
    end = atoi(expr+ 1);
  }
  else
  {
    start = end = atoi(expr);
    if ((q =  strchr(expr,'-')) && *(q+1) )
      end = atoi(q+1);
    else if (q && !*(q+1))
      end = 65535;
  }

  if (start < 1 || start > end)
    err_mesg(FATAL, "Your port specifications are illegal!");

  for(j=start; j <= end; j++) 
    ports[i++] = j;

  ports[i++] = 0;
  tmp = realloc(ports, i * sizeof(short));
  free(mem);
  
  return(tmp);
}

